/*+++++++++++++++++
  nmedris.c - input filter for Medline data. part of refdb
  markus@mhoenicka.de 8-13-00
  $Id: nmed2ris.c,v 1.14.2.1 2005/04/15 22:45:08 mhoenicka Exp $ 

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, see <http://www.gnu.org/licenses/>

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/* Todo: completely rewrite this mess along the lines of bib2ris */

/* This code is perhaps the most amusing in the whole refdb package. It all started ca. 1994 as a M$ QuickC program when our only access to Medline was through a networked Mac (wow - a real network! Unthinkable those days with our Win31 boxes). The retrieval program (Knowledge Finder) would write the data on a PC disk - but with Mac line endings (CR). Although the references would be digestable for Reference Manager (version 4, afair), the line ends were not. Lacking any useful tool on Win31 (nooo, nothing like tr, sed or the like), I decided to hack a filter that did the conversion. I added some conversions on the fly so the filter wrote valid RIS stuff to an output file. As I just had started programming in C and did not have any idea how to implement buffered input (fgets simply refused to work with CR :-)), the very first version just read the data byte for byte from the disk. But after only 2 days or so of hacking we could import Medline data into our Reference Manager. I'm still surprised that the hard drive didn't break down.

Later I added buffered input (this worked with all line endings, be it Mac, Unix, or PC) and a nice drag&drop GUI. More input formats were added, and all was nice - until I started using Linux. Now the GUI was suddenly awkward because the conversion was soo much easier to do in a nice shell like bash, and a new application (refdb) required that the medline converter works with pipes. So I had to extract parts of the old code and paste it into a command line app framework.

Some of the code has not been cleaned up since 1994, so don't be surprised if you find sections of code which are awkward or otherwise strange. The support for Mac line endings has been dropped to simplify things. Use something like tr on a good operating system if you need to handle Mac files */

/* current limitations */
/* support for lines > 64k not tested */

/* temporary hack to include cgi features. This should be selectable via
   a configure switch */
#define REFDB_CGI 1

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#include <getopt.h>
#include <syslog.h>
#include <errno.h>
#include <limits.h>

#include "refdb.h" /* refdb-client.h needs this */
#include "connect.h"
#include "refdb-client.h" /* for sighandler() */
#include "linklist.h" 
#include "pref.h" /* depends on linklist.h */
#include "strfncs.h" /* for LOG_PRINT definition, mstr*() */
#include "cgi.h"
#include "nmed2ris.h"

/*+ this array will hold the user preferences +*/
Prefs prefs[10] = {
  {"medsource", ""},
  {"outfile", ""},
  {"outappend", ""},
  {"ignoretag", ""},
  {"logfile", ""},
  {"logdest", ""},
  {"loglevel", ""},
  {"refdblib", ""},
  {"autokill", ""},
  {"", ""}
};

/* these are the configurable variables with the compile-time defaults */
char med_source[PREFS_BUF_LEN] = ""; /*+ Medline source +*/
char out_file[PREFS_BUF_LEN] = ""; /*+ output file +*/
char out_append[PREFS_BUF_LEN] = "t"; /*+ default output file append option +*/
char input_file[PREFS_BUF_LEN] = ""; /*+ input file +*/
char log_file[PREFS_BUF_LEN] = "/var/log/nmed2ris.log"; /*+ default log file +*/
char log_dest[PREFS_BUF_LEN] = "1"; /*+ default log destination (0 = stderr, 1 = syslog, 2 = log_file +*/
char log_level[PREFS_BUF_LEN] = "6"; /*+ default level up to which messages are logged (0 through 7). -1 means no logging +*/
char refdblib[PREFS_BUF_LEN] = ""; /* the location of shareable RefDB files */
char autokill[PREFS_BUF_LEN] = "1800"; /*+ default time in seconds until the CGI app will kill itself +*/
char ignore_tag[PREFS_BUF_LEN] = "t"; /* ignore unknown tags */
char confdir[_POSIX_PATH_MAX+1] = ""; /* path to the config files */
char complete_check = 0;		// bit-field to check for minimum number of fields
char convert_check = 0;		// bit-field to check for type of conversion error
char* outbuffer; /* will hold the output */

// predefined error strings for error log file
char conv_err_title[] = "conversion error in TI or CM string: title may be corrupted\n";
char conv_err_source[] = "conversion error in SO string: journal, year, volume, page may be corrupted\n";
char conv_err_author[] = "conversion error in AU string: author(s) may be corrupted\n";
char conv_err_unknown[] = "conversion error: tag unknown, information may get lost\n";
char conv_type_added[] = "standard publication type added\n";
char conv_missing_title[] = "no title string detected\n";
char conv_missing_pubyear[] = "no publication year detected\n";
char conv_missing_journal[] = "no journal detected\n";
char conv_missing_volume[] = "no volume detected\n";
char conv_missing_page[] = "no starting page detected\n";
char conv_missing_author[] = "no authors detected\n";
char conv_missing_type[] = "no publication type detected\n";

int n_log_level = 6; /* numeric version of log_level */
int n_log_dest = 1; /* numeric version of log_dest */
int n_read_stdin = 0; /* if 1, data come in via stdin */
int n_broken_pipe = 0; /* if 1, a SIGPIPE was caught */
int n_incomplete_string = 0; /* if 1, fgets read an incomplete string */
int n_inbuffer_len; /* allocated size of inbuffer */
int n_outbuffer_len; /* allocated size of outbuffer */
int n_waitfortype = 1; /* 1 until TYPE field is found */
int n_ignore_tag = 1; /* if 1, will ignore unknown tags */
int n_cgi = 0; /* if 1, we run as a cgi app */
unsigned int n_autokill = 1800; /* numeric version of autokill */

FILE* fp_logfile = NULL; /* ptr to logfile structure */
FILE* fp_infile; /* ptr to input file structure */
FILE* fp_outfile; /* ptr to output file structure */


/* forward declarations of static funtions */
void alrmhandler(int sig);

/********************************************************************/
/* the one and only main() */
/********************************************************************/
int main(int argc, char** argv) {
  char out_mode[3] = "ab"; /* default: file output is append */
  char source_type[3]; /* will hold filename prefix to test for data source */
  char sources[6][3] = {"BM", "CC", "GM", "HG", "KF", "PM"};
  char* inbuffer; /* will hold a line of the input */
  char msgbuf[512]; /* buffer to assemble error messages */
  char* html_string;
  int n_readinit = 1; /* if 1, init file will be read */
  int i, j; /* guess what */
  int n_append = 1; /* if 1, output will be appended to existing file. If 0, output overwrites existing file */
  int n_source_type = PUBMED; /* type of input source, set to default */
  int n_fixed_source = 0; /* if one, don't use filename prefixes */
  int n_opt; /* used by getopt() */
  int n_use_defaultfilter = 1; /* will be set to 0 if correct filter is specified */
  int n_content_length;
  struct sigaction act, oldact, alrm, oldalrm; /* structure for signal handler */

  char delayed_cgi_errmsg[64] = "";
  char* cgi_data;
  char* request_method;
  char* content_length;
  FILE* errstream;
#ifdef REFDB_CGI
  struct liliform sentinelili;
  struct liliform* ptr_current;
#endif

  /* initialize signal handler */
  n_broken_pipe = 0;
  act.sa_handler = pipehandler;
  sigemptyset(&act.sa_mask);
  act.sa_flags = 0;
  sigaction(SIGPIPE, &act, 0);

  /* test how we were invoked */
#ifdef REFDB_CGI
  n_cgi = is_cgi(&request_method, &content_length, delayed_cgi_errmsg); /* 1 if we are running as a cgi app */
#endif

  /* redirect error messages correctly */
  if (n_cgi) {
    errstream = stdout;
  }
  else {
    errstream = stderr;
  }

  if (!isatty(fileno(stdin))) { /* if we receive data on stdin via a redirection or a pipe */
    n_read_stdin = 1;
  }
    
#ifdef REFDB_CGI
  if (n_cgi) {
    alrm.sa_handler = alrmhandler;
    sigemptyset(&alrm.sa_mask);
    alrm.sa_flags = 0;

    if (sigaction(SIGALRM, &alrm, &oldalrm) != 0) {
      write_err("initializing signal handler failed\n");
      LOG_PRINT(LOG_ERR, "could not initialize signal handler");
      exit(1);
    }
  }
#endif

  prefs[0].varvalue = med_source;
  prefs[1].varvalue = out_file;
  prefs[2].varvalue = out_append;
  prefs[3].varvalue = ignore_tag;
  prefs[4].varvalue = log_file;
  prefs[5].varvalue = log_dest;
  prefs[6].varvalue = log_level;
  prefs[7].varvalue = refdblib;

  /* look for custom config directory */
  for (i = 0; i < argc; i++) {
    if (argv[i][0] == '-' && argv[i][1] == 'y') {
      strncpy(confdir, argv[i+1], _POSIX_PATH_MAX);
      confdir[_POSIX_PATH_MAX] = '\0';
      break;
    }
  }

  if (!n_cgi) {
    /* a slimy hack to detect the -q option before we run getopt */
    for (i = 0; i < argc; i++) {
      if (argv[i][0] == '-' && argv[i][1] == 'q') {
	n_readinit = 0;
	break;
      }
    }

    if (n_readinit) {
      /* read config file settings */
      read_prefs(prefs, "nmed2risrc", 0);
    }

    /* read command line settings. These may override the config file settings */
    while ((n_opt = getopt(argc, argv, "e:f:hi:l:L:o:O:qs:vy:")) != -1) {
      switch (n_opt) {
      case 'e':
	strncpy(log_dest, optarg, PREFS_BUF_LEN);
	log_dest[PREFS_BUF_LEN-1] = '\0';
	break;
      case 'f':
	strncpy(input_file, optarg, PREFS_BUF_LEN);
	input_file[PREFS_BUF_LEN-1] = '\0';
	break;
      case 'h':
	fprintf(stderr, "nmed2ris converts references from various online sources to RIS format as understood by the refdb package.\n");
	fprintf(stderr, "Usage: nmed2ris [-e destination] [-f input-file] [-h] [-i t|f] [-l log-level] [-L logfile] [-o output_file] [-O output_file] [-s data_source] [-v] [-y confdir]\nOptions: -e set log destination (0 stderr;1 syslog;2 logfile)\n         -f read input from file\n         -h print this help and exit\n         -i if t, ignore unknown tags\n         -l set the log level\n         -L set the log file\n         -o send output to file (overwrite)\n         -O send output to file (append)\n         -s specify data source\n         -v print version info and exit\n         -y look for configuration files in confdir\n");
	exit (0);
	break;
      case 'i':
	if (*optarg == 't') {
	  n_ignore_tag = 1;
	}
	else {
	  n_ignore_tag = 0;
	}
	break;
      case 'l':
	strncpy(log_level, optarg, PREFS_BUF_LEN);
	log_level[PREFS_BUF_LEN-1] = '\0';
	break;
      case 'L':
	strncpy(log_file, optarg, PREFS_BUF_LEN);
	log_file[PREFS_BUF_LEN-1] = '\0';
	break;
      case 'o':
	strcpy(out_mode, "wb");
	/* fall through */
      case 'O':
	/* default is out_mode = ab */
	strncpy(out_file, optarg, PREFS_BUF_LEN);
	out_file[PREFS_BUF_LEN-1] = '\0';
	break;
      case 'q':
	/* do nothing; prevents message about unknown -q option */
	break;
      case 's':
	strncpy(med_source, optarg, PREFS_BUF_LEN);
	med_source[PREFS_BUF_LEN-1] = '\0';
	break;
      case 'v':
	printf("nmed2ris %s\nnmed2ris is released under the GNU General Public License. You may give away and alter this software with a few restrictions. See the file COPYING that accompanied this program for the fine print.\n", VERSION);
	break;
      case 'y':
	/* do nothing, this option is used before getopt runs */
	break;
      case ':':
	fprintf(stderr, "Usage: nmed2ris [-e destination] [-f input-file] [-h] [-i t|f] [-l log-level] [-L logfile] [-o output_file] [-O output_file] [-s data_source] [-v] [-y confdir]\nOptions: -e set log destination (0 stderr;1 syslog;2 logfile)\n         -f read input from file\n         -h print this help and exit\n         -i if t, ignore unknown tags\n         -l set the log level\n         -L set the log file\n         -o send output to file (overwrite)\n         -O send output to file (append)\n         -s specify data source\n         -v print version info and exit\n         -y look for configuration files in confdir\n");
	exit (1);
	break;
      case '?':
	fprintf(stderr, "unknown option %c: use nmed2ris -h to display usage\n", optopt);
	exit (1);
	break;
      }
    }
  }
  else { /* if we run as CGI app */
    read_prefs((void*)prefs, "nmed2riscgirc", 0);
  }

#ifdef REFDB_CGI
  if (n_cgi) { /* no killer alarm in interactive mode */
    n_autokill = (unsigned int)atoi(autokill);
  
    /* this alarm will go off after n_autokill seconds and cause the CGI app
       to exit as it is apparently hanging */
    if (n_autokill) {
      alarm(n_autokill);
    }
  }
#endif

  /* make some conversions */
  n_log_level = num_loglevel(log_level);
  n_log_dest = num_logdest(log_dest);

  if (strcmp(ignore_tag, "f") == 0) {
    n_ignore_tag = 0;
  }

  /* start logging */
  if (n_log_dest == 2) { /* use custom log file */
    if ((fp_logfile = fopen(log_file, "ab")) == NULL) {
      n_log_dest = 1; /* fall back to syslog */
      LOG_PRINT(LOG_WARNING, "could not open custom log file");
    }
  }

  LOG_PRINT(LOG_INFO, "nmed2ris started");

  /* now report incorrect cgi input, if any */
  if (delayed_cgi_errmsg[0]) {
    LOG_PRINT(LOG_ERR, delayed_cgi_errmsg);
    exit(1);
  }

  /* find out what type of input we have */
  if (*med_source) { /* if set on the command line or in the init file */
    strup(med_source); /* to uppercase */
    n_fixed_source = 1; /* don't use filename prefixes */
    for (i = 0; i < 6; i++) {
      if (strcmp(med_source, sources[i]) == 0) {
	n_source_type = i;
	n_use_defaultfilter = 0;
	break;
      }
    }
/*      if (strcmp(med_source, "BM") == 0) { */
/*        n_source_type = BIOMEDNET; */
/*      } */
/*      if (strcmp(med_source, "CC") == 0) { */
/*        n_source_type = CCOD; */
/*      } */
/*      if (strcmp(med_source, "GM") == 0) { */
/*        n_source_type = GRATEMED; */
/*      } */
/*      if (strcmp(med_source, "HG") == 0) { */
/*        n_source_type = HEALTHGATE; */
/*      } */
/*      if (strcmp(med_source, "KF") == 0) { */
/*        n_source_type = KNOWFINDER; */
/*      } */
/*      if (strcmp(med_source, "PM") == 0) { */
/*        n_source_type = PUBMED; */
/*      } */
  }

  /* open the output file if needed */
  if (*out_file) {
    if ((fp_outfile = fopen(out_file, out_mode)) == NULL) {
      sprintf(msgbuf, "could not open output file %s. nmed2ris stopped.", argv[i]);
      LOG_PRINT(LOG_CRIT, msgbuf);
      exit (1);
    }
  }
  else {
    fp_outfile = stdout;
  }

  /* get us some memory for the input */
  n_inbuffer_len = INFILE_CHUNK_SIZE; /* something to start with */
  if ((inbuffer = (char*)malloc(n_inbuffer_len)) == NULL) {
    LOG_PRINT(LOG_CRIT, "out of memory. nmed2ris stopped.");
    exit (1);
  }

  /* ...and for the output */
  n_outbuffer_len = INFILE_CHUNK_SIZE; /* something to start with */
  if ((outbuffer = (char*)malloc(n_outbuffer_len)) == NULL) {
    free(inbuffer);
    LOG_PRINT(LOG_CRIT, "out of memory. nmed2ris stopped.");
    exit (1);
  }

  inbuffer[0] = '\0'; /* terminate string */
  outbuffer[0] = '\0'; /* terminate string */

  if (n_cgi) { /* we're running as a cgi app */
#ifdef REFDB_CGI
    /* initialize linked list */
    sentinelili.name[0] = '\0';
    sentinelili.value = NULL;
    sentinelili.ptr_next = NULL;

    /* read string from stdin and chop it */
    n_content_length = atoi(content_length);
  
    cgi_data = (char*)malloc(n_content_length + 1);
    if (!cgi_data) {
      cgi_header(CGI_PLAIN);
      fprintf(errstream, "out of memory\n");
      LOG_PRINT(LOG_WARNING, "out of memory");
      exit(1);
    }

    if (fread(cgi_data, 1, n_content_length, stdin) != n_content_length) {
      cgi_header(CGI_PLAIN);
      fprintf(errstream, "could not read cgi data\n");
      free(cgi_data);
      LOG_PRINT(LOG_ERR, "could not read cgi data");
      exit(1);
    }

    cgi_data[n_content_length] = '\0'; /* convert to a C-style string */
    LOG_PRINT(LOG_DEBUG, cgi_data);

    if (decode_cgi(cgi_data, &sentinelili)) {
      delete_all_liliform(&sentinelili);
      free(cgi_data);
      cgi_header(CGI_PLAIN);
      fprintf(errstream, "could not decode cgi data\n");
      LOG_PRINT(LOG_ERR, "could not decode cgi data");
      exit(1);
    }

    free(cgi_data);

    /* deduce command */
    if ((ptr_current = get_liliform(&sentinelili, "addref")) != NULL) {
      cgi_header(CGI_HTML);
      LOG_PRINT(LOG_DEBUG, ptr_current->value);
      if ((html_string = load_html("refdbadd_head", refdblib)) == NULL) {
	LOG_PRINT(LOG_CRIT, "failed to load template");
	printf("failed to load template\n");
	return 1;
      }
      else {
	printf(html_string);
	free(html_string);
      }

      if (convert_data(&(ptr_current->value), n_source_type)) {
	LOG_PRINT(LOG_ERR, "errors encountered while processing CGI data");
	/*        n_gen_error = 1; */
      }
      if ((html_string = load_html("refdbadd_foot", refdblib)) == NULL) {
	LOG_PRINT(LOG_CRIT, "failed to load template");
	printf("failed to load template\n");
	return 1;
      }
      else {
	printf(html_string);
	free(html_string);
      }

    }
    else {
      delete_all_liliform(&sentinelili);
      cgi_header(CGI_PLAIN);
      fprintf(errstream, "unknown command\n");
      LOG_PRINT(LOG_ERR, "unknown command");
      exit(1);
    }

    /* clean up */
    delete_all_liliform(&sentinelili);
#endif
  }
  else if (n_read_stdin) { /* read data from stdin */
    fp_infile = stdin;
    if (n_use_defaultfilter) {
      LOG_PRINT(LOG_WARNING, "read data from stdin. no source type given, use default");
    }
    else {
      sprintf(msgbuf, "read data from stdin. use %s filter", sources[n_source_type]);
      LOG_PRINT(LOG_INFO, msgbuf);
    }
    convert_data(&inbuffer, n_source_type);
  }
  else { /* read data from all files on the command line */
    for (i = optind; i < argc; i++) { /* loop over all filenames */
      if (!n_fixed_source) {
	n_use_defaultfilter = 1;
	strncpy(source_type, argv[i], 2); /* get prefix */
	strup(source_type); /* make uppercase */
	for (j = 0; j < 6; j++) {
	  if (strcmp(source_type, sources[j]) == 0) {
	    n_source_type = j;
	    n_use_defaultfilter = 0;
	    break;
	  }
	}
/*  	if (strcmp(source_type, "BM") == 0) { */
/*  	  n_source_type = BIOMEDNET; */
/*  	} */
/*  	else if (strcmp(source_type, "CC") == 0) { */
/*  	  n_source_type = CCOD; */
/*  	} */
/*  	else if (strcmp(source_type, "GM") == 0) { */
/*  	  n_source_type = GRATEMED; */
/*  	} */
/*  	else if (strcmp(source_type, "HG") == 0) { */
/*  	  n_source_type = HEALTHGATE; */
/*  	} */
/*  	else if (strcmp(source_type, "KF") == 0) { */
/*  	  n_source_type = KNOWFINDER; */
/*  	} */
/*  	else if (strcmp(source_type, "PM") == 0) { */
/*  	  n_source_type = PUBMED; */
/*  	} */
      }
      if ((fp_infile = fopen(argv[i], "rb")) == NULL) {
	sprintf(msgbuf, "could not open input file %s", argv[i]);
	LOG_PRINT(LOG_WARNING, msgbuf);
	/* try next file */
      }
      else {
	if (n_use_defaultfilter) {
	  sprintf(msgbuf, "convert %s. use default filter", argv[i]);
	}
	else {
	  sprintf(msgbuf, "convert %s. use %s filter", argv[i], sources[n_source_type]);
	}
	LOG_PRINT(LOG_INFO, msgbuf);
	if (!convert_data(&inbuffer, n_source_type)) {
/*  	  printf("convert_data returns 0\n"); */
	  sprintf(msgbuf, "errors encountered while converting %s", argv[i]);
	  LOG_PRINT(LOG_ERR, msgbuf);
	}
	fclose(fp_infile);
      }

    }
  }
  free(inbuffer);
  free(outbuffer);

  LOG_PRINT(LOG_INFO, "nmed2ris stopped");

  if (fp_logfile) {
    fclose(fp_logfile); /* not strictly necessary, but more polite */
  }
  exit (0);
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  write_out(): writes a buffer to the output stream or into an
               output buffer

  int write_out 1 if successful, 0 if error 

  char *inbuffer a buffer for the incoming data

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int write_out(char* inbuffer) {
  char* new_outbuffer;
  size_t len;

  if (n_waitfortype) {
    /* write to an output buffer */
    if ((new_outbuffer = mstrcat(outbuffer, inbuffer, &n_outbuffer_len, 0)) == NULL) {
      LOG_PRINT(LOG_WARNING, "memory allocation failed");
      return 0;
    }
    else {
      outbuffer = new_outbuffer;
    }
  }
  else {
    /* write to output stream */
    len = strlen(inbuffer);
    if (fwrite(inbuffer, sizeof(char), len, fp_outfile) != len) {
      LOG_PRINT(LOG_WARNING, "writing output failed");
      return 0;
    }
  }
  return 1;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  search_year(): searches for 4-digit year in a string and terminates
                 the string after the year

  char* search_year ptr to year if successful, NULL if no year found

  char *token ptr to the buffer containing the string to examine

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
char* search_year(char* token) {
  size_t i;
  char* digit = NULL;
	
  for (i = 0; i < strlen(token)-3; i++) { /* stop searching at omega-3 to save time */
    if (isdigit((int)token[i])) { /* if there's a digit in the string */
      /* check the next three characters */
      if (isdigit((int)token[i+1]) && isdigit((int)token[i+2]) &&
	  isdigit((int)token[i+3])) {
	token[i+4] = '\0'; /* if four-digit number found, delimit string */
	digit = &token[i]; /* calculate pointer to begin of digit string */
	break; /* and jump out of for loop */
      }
    }
  }
  return digit; /* will still be null if no year found */
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  write_source(): parses the source information and writes output

  int write_source  1 if successful, 0 if error

  char *inbuffer ptr to the buffer containing the string to examine

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int write_source(char* inbuffer) {
  int retvalue = 0;
  int success = 1;
  char part[256] = "";
  char *token;
  char *token_py;
  char *token1;
  char *token2;
  char *end_token;

  /* look for journal and year */
  end_token = strchr(&inbuffer[6], (int)';');
  if (end_token != NULL) {
    *end_token = '\0'; /* discard separator ";" */
    token_py = search_year(&inbuffer[6]); /* search the year substring */
    if (token_py != NULL) { /* if there is a year */
      *(token_py-1) = '\0'; /* delimit journal */
      stripwhite(&inbuffer[6], 2, 1); /* remove trailing space at end of journal */

      if (!write_out("PY  - ")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0; /* error */
      }
      if (!write_out(token_py)) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0; /* error */
      }
      if (&token_py[4] != end_token) { /* if there is some other info */
	if (!write_out("///")) {
	  LOG_PRINT(LOG_WARNING, "writing output failed");
	  return 0; /* error */
	}
	if (!write_out(&token_py[5])) { /* other info */
	  LOG_PRINT(LOG_WARNING, "writing output failed");
	  return 0; /* error */
	}
      }
      if (!write_out("\n")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0; /* error */
      }
	
      complete_check |= 2;
    }
    else {
      convert_check |= 2;
      return 0;
    }

    if (!write_out("JA  - ")) { /* put journal abbrev into JA field */
      LOG_PRINT(LOG_WARNING, "writing output failed");
      return 0; /* error */
    }
    if (!write_out(&inbuffer[6])) {
      LOG_PRINT(LOG_WARNING, "writing output failed");
      return 0; /* error */
    }
    if (!write_out("\n")) {
      LOG_PRINT(LOG_WARNING, "writing output failed");
      return 0; /* error */
    }
    complete_check |= 4;			
  }
  else {
    convert_check |= 2;		
    return 0;
  }
    
  /* look for volume */
  token = end_token + 1;
  end_token = strchr(token, (int)':');
/*    printf("end_token:%s<<\n", end_token); */
  if (end_token != NULL) {
    token1 = strchr(token+1, (int)'('); /* search for issue */
    if (token1 != NULL) {
      *token1 = '\0';
      if (!write_out("VL  - ")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0; /* error */
      }
      if (!write_out(token)) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0; /* error */
      }
      if (!write_out("\n")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0; /* error */
      }
      complete_check |= 8;
				
      token2 = strchr(++token1, ')'); /* set token1 to start of issue token */
      if (token2 != NULL) {
	*token2 = '\0';
      }
      if (!write_out("IS  - ")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0; /* error */
      }
      if (!write_out(token1)) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0; /* error */
      }
      if (!write_out("\n")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0; /* error */
      }
    }
    else {
      token1 = strstr(token+1, "Pt"); /* some dont have the volume(issue) notation */
      if (token1 != NULL) {
	*token1 = '\0';
	if (sscanf(token1+3, "%[0-9]", part) != 0) {
	  if (!write_out("IS  - ")) {
	    LOG_PRINT(LOG_WARNING, "writing output failed");
	    return 0; /* error */
	  }
	  if (!write_out(part)) {
	    LOG_PRINT(LOG_WARNING, "writing output failed");
	    return 0; /* error */
	  }
	  if (!write_out("\n")) {
	    LOG_PRINT(LOG_WARNING, "writing output failed");
	    return 0; /* error */
	  }
	}
      }
      else { /* no issue or part information */
	*end_token = '\0';
      }

      if (!write_out("VL  - ")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0; /* error */
      }
      if (!write_out(token)) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0; /* error */
      }
      if (!write_out("\n")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0; /* error */
      }
      complete_check |= 8;
    }
  }
  else if (*token) {
    token1 = strchr(token+1, (int)'('); /* search for issue */
    if (token1 != NULL) {
      *token1 = '\0';
      if (!write_out("VL  - ")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0; /* error */
      }
      if (!write_out(token)) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0; /* error */
      }
      if (!write_out("\n")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0; /* error */
      }
      complete_check |= 8;
				
      token2 = strchr(++token1, ')'); /* set token1 to start of issue token */
      if (token2 != NULL) {
	*token2 = '\0';
      }
      if (!write_out("IS  - ")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0; /* error */
      }
      if (!write_out(token1)) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0; /* error */
      }
      if (!write_out("\n")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0; /* error */
      }
    }
    else {
      token1 = strstr(token+1, "Pt"); /* some dont have the volume(issue) notation */
      if (token1 != NULL) {
	*token1 = '\0';
	if (sscanf(token1+3, "%[0-9]", part) != 0) {
	  if (!write_out("IS  - ")) {
	    LOG_PRINT(LOG_WARNING, "writing output failed");
	    return 0; /* error */
	  }
	  if (!write_out(part)) {
	    LOG_PRINT(LOG_WARNING, "writing output failed");
	    return 0; /* error */
	  }
	  if (!write_out("\n")) {
	    LOG_PRINT(LOG_WARNING, "writing output failed");
	    return 0; /* error */
	  }
	}
      }

      if (!write_out("VL  - ")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0; /* error */
      }
      if (!write_out(token)) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0; /* error */
      }
      if (!write_out("\n")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0; /* error */
      }
      complete_check |= 8;
    }
    return 1;
  }
  else {
    convert_check |= 2;		
    return 0;
  }
		
  /* look for start page */
  token = strchr(end_token+1, (int)'-');
  if (token != NULL) {
    *token = '\0';
    if (!write_out("SP  - ")) {
      LOG_PRINT(LOG_WARNING, "writing output failed");
      return 0; /* error */
    }

    /* eliminate the trailing period that PubMed emits since approx Jan 2001 */
    if (*(end_token+strlen(end_token+1)) == '.') {
      *(end_token+strlen(end_token+1)) = '\0';
    }

    if (!write_out(end_token+1)) {
      LOG_PRINT(LOG_WARNING, "writing output failed");
      return 0; /* error */
    }
    if (!write_out("\n")) {
      LOG_PRINT(LOG_WARNING, "writing output failed");
      return 0; /* error */
    }
    complete_check |= 16;
  }
  else if (*(end_token + 1)) {
    if (!write_out("SP  - ")) {
      LOG_PRINT(LOG_WARNING, "writing output failed");
      return 0; /* error */
    }

    /* eliminate the trailing period that PubMed emits since approx Jan 2001 */
    if (*(end_token+strlen(end_token+1)) == '.') {
      *(end_token+strlen(end_token+1)) = '\0';
    }

    if (!write_out(end_token+1)) {
      LOG_PRINT(LOG_WARNING, "writing output failed");
      return 0; /* error */
    }
    if (!write_out("\n")) {
      LOG_PRINT(LOG_WARNING, "writing output failed");
      return 0; /* error */
    }
    complete_check |= 16;
    return 1;
  }
  else {
    convert_check |= 2;		
    return 0;
  }
    
/* look for end page (no prob if missing) */
  if (*(token+1) != '\0') {
    if (!write_out("EP  - ")) {
      LOG_PRINT(LOG_WARNING, "writing output failed");
      return 0; /* error */
    }

    /* eliminate the trailing period that PubMed emits since approx Jan 2001 */
    if (*(token+strlen(token+1)) == '.') {
      *(token+strlen(token+1)) = '\0';
    }

    if (!write_out(token+1)) {
      LOG_PRINT(LOG_WARNING, "writing output failed");
      return 0; /* error */
    }
    if (!write_out("\n")) {
      LOG_PRINT(LOG_WARNING, "writing output failed");
      return 0; /* error */
    }
  }
		
  return 1;
}
	
  
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  convert_author(): converts medline AU field into RIS AU field

  int convert_author 1 if successful, 0 if error 

  char *inbuffer a buffer for the incoming data

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int convert_author(char* inbuffer) {
  char *space;
  char *space1;
  int num_initial = 0;
  int i;
  char initial_string[] = " .";
  int name_suf = 0; /* if 1, the name has a suffix like Jr. */
	
  if (strlen(inbuffer) == 0) { /* take shortcut if string is empty */
    return 1; /* an empty string is ok, but useless */
  }
  if (!write_out("AU  - ")) {
    LOG_PRINT(LOG_WARNING, "writing output failed");
    return 0;
  }
	
  if (strcmp(inbuffer, "et al") == 0) { /* if it's an et al author */
    if (!write_out("et al.\n")) {
      LOG_PRINT(LOG_WARNING, "writing output failed");
      return 0;
    }
    return 1; /* finished */
  }
	
  space = strrchr(inbuffer, 32); /* search for rightmost space separating lastname and initials */
	
  if (space != NULL) { /* if there's a space */
    *space = '\0'; /* replace it with a NULL */
    num_initial = strlen(&space[1]); /* calculate number of initials */
    space1 = strrchr(inbuffer, 32); /* search for something like Jr */
    if (space1 != NULL && isupper((int)space[-1])) {
      *space1 = '\0';
      num_initial = strlen(&space1[1]); /* calculate number of initials */
      name_suf = 1;
    }
    else {
      space1 = space;
    }
  }
  if (!write_out(inbuffer)) { /* add the lastname */
    LOG_PRINT(LOG_WARNING, "writing output failed");
    return 0;
  } 
  if (!write_out(",")) { /* and separate it with "," from the rest */
    LOG_PRINT(LOG_WARNING, "writing output failed");
    return 0;
  } 

  for (i = 0; i < num_initial; i++) { /* go through all initials */
    *initial_string = space1[i+1];
    if (!write_out(initial_string)) {
      LOG_PRINT(LOG_WARNING, "writing output failed");
      return 0;
    }
  }
  if (name_suf) {
    if (!write_out(",")) {
      LOG_PRINT(LOG_WARNING, "writing output failed");
      return 0;
    }
    if (!write_out(&space[1])) {
      LOG_PRINT(LOG_WARNING, "writing output failed");
      return 0;
    }
    if (!write_out(".")) {
      LOG_PRINT(LOG_WARNING, "writing output failed");
      return 0;
    }
  } 
  if (!write_out("\n")) {
    LOG_PRINT(LOG_WARNING, "writing output failed");
    return 0;
  }
	
  return 1;
} 

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  add_subheading(): adds the subheadings in correct format to a heading

  int add_subheading 1 if successful, 0 if error 

  char *inbuffer a buffer for the incoming data

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int add_subheading(char* inbuffer) {
  int nretval = 1;
	
  switch (inbuffer[0]) {
  case 'A':
    switch (inbuffer[1]) {
    case 'A':
      if (!write_out(" [Analogs & Derivatives]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'B':
      if (!write_out(" [Abnormalities]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'D':
      if (!write_out(" [Administration & Dosage]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'E':
      if (!write_out(" [Adverse Effects]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'G':
      if (!write_out(" [Agonists]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'H':
      if (!write_out(" [Anatomy & Histology]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'I':
      if (!write_out(" [Antagonists & Inhibitors]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'N':
      if (!write_out(" [Analysis]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    default:
      nretval = 0;
      break;
    }
    break;
			
  case 'B':
    switch (inbuffer[1]) {
    case 'I':
      if (!write_out(" [Biosynthesis]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'L':
      if (!write_out(" [Blood]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'S':
      if (!write_out(" [Blood Supply]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    default:
      nretval = 0;
      break;
    }
    break;
			
  case 'C':
    switch (inbuffer[1]) {
    case 'F':
      if (!write_out(" [Cerebrospinal Fluid]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'H':
      if (!write_out(" [Chemistry]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'I':
      if (!write_out(" [Chemically Induced]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'L':
      if (!write_out(" [Classification]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'N':
      if (!write_out(" [Congenital]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'O':
      if (!write_out(" [Complications]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'S':
      if (!write_out(" [Chemical Synthesis]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'T':
      if (!write_out(" [Contraindications]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'Y':
      if (!write_out(" [Cytology]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    default:
      nretval = 0;
      break;
    }
    break;
			
  case 'D':
    switch (inbuffer[1]) {
    case 'E':
      if (!write_out(" [Drug Effects]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'F':
      if (!write_out(" [Deficiency]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'H':
      if (!write_out(" [Diet Therapy]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'I':
      if (!write_out(" [Diagnosis]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'T':
      if (!write_out(" [Drug Therapy]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'U':
      if (!write_out(" [Diagnostic Use]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    default:
      nretval = 0;
      break;
    }
    break;
			
  case 'E':
    switch (inbuffer[1]) {
    case 'C':
      if (!write_out(" [Economics]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'D':
      if (!write_out(" [Education]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'H':
      if (!write_out(" [Ethnology]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'M':
      if (!write_out(" [Embryology]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'N':
      if (!write_out(" [Enzymology]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'P':
      if (!write_out(" [Epidemiology]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'T':
      if (!write_out(" [Etiology]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    default:
      nretval = 0;
      break;
    }
    break;
			
  case 'G':
    switch (inbuffer[1]) {
    case 'D':
      if (!write_out(" [Growth & Development]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'E':
      if (!write_out(" [Genetics]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    default:
      nretval = 0;
      break;
    }
    break;
			
  case 'H':
    switch (inbuffer[1]) {
    case 'I':
      if (!write_out(" [History]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    default:
      nretval = 0;
      break;
    }
    break;
			
  case 'I':
    switch (inbuffer[1]) {
    case 'M':
      if (!write_out(" [Immunology]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'N':
      if (!write_out(" [Injuries]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'P':
      if (!write_out(" [Isolation & Purification]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'R':
      if (!write_out(" [Innervation]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'S':
      if (!write_out(" [Instrumentation]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    default:
      nretval = 0;
      break;
    }
    break;
			
  case 'L':
    switch (inbuffer[1]) {
    case 'J':
      if (!write_out(" [Legislation & Jurisprudence]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    default:
      nretval = 0;
      break;
    }
    break;
			
  case 'M':
    switch (inbuffer[1]) {
    case 'A':
      if (!write_out(" [Manpower]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'E':
      if (!write_out(" [Metabolism]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'I':
      if (!write_out(" [Microbiology]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'O':
      if (!write_out(" [Mortality]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'T':
      if (!write_out(" [Methods]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    default:
      nretval = 0;
      break;
    }
    break;
			
  case 'N':
    switch (inbuffer[1]) {
    case 'U':
      if (!write_out(" [Nursing]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    default:
      nretval = 0;
      break;
    }
    break;
			
  case 'O':
    switch (inbuffer[1]) {
    case 'G':
      if (!write_out(" [Organization & Administration]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    default:
      nretval = 0;
      break;
    }
    break;
			
  case 'P':
    switch (inbuffer[1]) {
    case 'A':
      if (!write_out(" [Pathology]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'C':
      if (!write_out(" [Prevention & Control]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'D':
      if (!write_out(" [Pharmacology]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'H':
      if (!write_out(" [Physiology]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'K':
      if (!write_out(" [Pharmacokinetics]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'O':
      if (!write_out(" [Poisoning]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'P':
      if (!write_out(" [Physiopathology]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'S':
      if (!write_out(" [Parasitology]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'X':
      if (!write_out(" [Psychology]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'Y':
      if (!write_out(" [Pathogenicity]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    default:
      nretval = 0;
      break;
    }
    break;
			
  case 'R':
    switch (inbuffer[1]) {
    case 'A':
      if (!write_out(" [Radiography]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'E':
      if (!write_out(" [Radiation Effects]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'H':
      if (!write_out(" [Rehabilitation]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'I':
      if (!write_out(" [Radionuclide Imaging]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'T':
      if (!write_out(" [Radiotherapy]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    default:
      nretval = 0;
      break;
    }
    break;
			
  case 'S':
    switch (inbuffer[1]) {
    case 'C':
      if (!write_out(" [Secondary]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'D':
      if (!write_out(" [Supply & Distribution]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'E':
      if (!write_out(" [Secretion]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'N':
      if (!write_out(" [Statistics & Numerical Data]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'T':
      if (!write_out(" [Standards]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'U':
      if (!write_out(" [Surgery]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    default:
      nretval = 0;
      break;
    }
    break;
			
  case 'T':
    switch (inbuffer[1]) {
    case 'D':
      if (!write_out(" [Trends]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'H':
      if (!write_out(" [Therapy]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'M':
      if (!write_out(" [Transmission]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'O':
      if (!write_out(" [Toxicity]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'R':
      if (!write_out(" [Transplantation]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'U':
      if (!write_out(" [Therapeutic Use]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    default:
      nretval = 0;
      break;
    }
    break;
			
  case 'U':
    switch (inbuffer[1]) {
    case 'L':
      if (!write_out(" [Ultrastructure]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'R':
      if (!write_out(" [Urine]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'S':
      if (!write_out(" [Ultrasonography]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'T':
      if (!write_out(" [Utilization]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    default:
      nretval = 0;
      break;
    }
    break;
			
  case 'V':
    switch (inbuffer[1]) {
    case 'E':
      if (!write_out(" [Veterinary]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    case 'I':
      if (!write_out(" [Virology]")) {
	LOG_PRINT(LOG_WARNING, "writing output failed");
	return 0;
      }
      break;
    default:
      nretval = 0;
      break;
    }
    break;
			
  default:
    nretval = 0;
    break;
  }
  return nretval;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  convert_data(): converts references to RIS format

  int convert_data 1 if successful, 0 if error 

  char **inbuffer ptr to buffer for the incoming data. If the buffer is not
                 empty, the fn will use the data in this string.
		 Otherwise it will read the data from fp_infile.

  int n_source_type specifies the Medline source (see values in .h)

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int convert_data(char** inbuffer, int n_source_type) {
  char ref_type[8];
  char* read_result; /* return value of fgets */
  char* new_outbuffer; /* for realloc()ing outbuffer */
  char message_buffer[512]; /* for assembling log messages */
  int n_inbuf_len; /* length of string in inbuffer */
  int n_eof = 0; /* indicates end-of-file condition */
  int n_total_refs = 0; /* total number of references */
  int n_ref_counter = 0; /* number of references */
  int n_error_counter = 0; /* number of errors */
  int n_warning_counter = 0; /* number of warnings */
  int n_havedata = 0; /* set to 1 if inbuffer contains data */
  struct convstate cstate;

  /* initialize struct */
  cstate.n_convert_error = 0;
  cstate.n_outomem = 0;
  cstate.n_outerror = 0;
  cstate.n_read_err = 0;
  cstate.n_read_incomplete = 0;

  if (**inbuffer) {
    n_havedata = 1;
  }

  /* convert references main loop - loop until no more lines to read */
  while (!n_eof) {
    /* skip leading empty lines */
    if (!n_havedata) {
      n_eof = skip_lines_fp(*inbuffer, fp_infile, &cstate);
    }
    else {
      n_eof = skip_lines_s(*inbuffer);
    }

    if (n_eof) {
      break; /* we're done */
    }

    n_ref_counter++;
    sprintf(message_buffer, "start converting reference %d", n_ref_counter);
    LOG_PRINT(LOG_INFO, message_buffer);
			
    /* start reference, depending on input type */
    if (n_source_type != BIOMEDNET) { /* if not BioMedNet input format */
      /* start output reference with a blank line */
      if (!n_havedata) {
	if (fwrite("\nTY  - JOUR\n", sizeof(char), 12, fp_outfile) < 12) {
	  LOG_PRINT(LOG_WARNING, "writing output failed");
	  cstate.n_outerror = 1;
	  n_waitfortype = 0;
	}
      }
      else {
	/* we need two linebreaks as browsers tend to swallow one */
	if (fwrite("\n\nTY  - JOUR\n", sizeof(char), 13, fp_outfile) < 13) {
	  LOG_PRINT(LOG_WARNING, "writing output failed");
	  cstate.n_outerror = 1;
	  n_waitfortype = 0;
	}
      }
    }

    /* read items of reference */
    if (!n_havedata) {
      n_eof = convert_ref_fp(inbuffer, &n_waitfortype, n_source_type, &cstate);
    }
    else {
      n_eof = convert_ref_s(inbuffer, &n_waitfortype, n_source_type, &cstate);
    }

    /* ToDo: check for errors after break */

    if (!(complete_check & 32)) {   /* if missing author */
      if ((new_outbuffer = mstrcat(outbuffer, "AU  - Anonymous\r\n", &n_outbuffer_len, 0)) == NULL) { /* add author string */
	cstate.n_outomem = 1;
	break;
      }
      else {
	outbuffer = new_outbuffer;
      }
      complete_check |= 32;   /* missing author will be no error */
    }
			
    if (cstate.n_convert_error || complete_check != 127) {   /* if conversion error, dump to error log file */
      sprintf(message_buffer, "Reference: %d/%d\n", n_ref_counter, n_total_refs);      /* indicate ref number */
      LOG_PRINT(LOG_INFO, message_buffer);
					
      /* indicate type of error */
      if (cstate.n_convert_error) {
	if (convert_check & 1) {
	  n_error_counter++;
	  LOG_PRINT(LOG_WARNING, conv_err_title);
	}
	if (convert_check & 2) {
	  n_error_counter++;
	  LOG_PRINT(LOG_WARNING, conv_err_source);
	}
	if (convert_check & 4) {
	  n_error_counter++;
	  LOG_PRINT(LOG_WARNING, conv_err_author);
	}
	if (convert_check & 8) {
	  n_error_counter++;
	  LOG_PRINT(LOG_WARNING, conv_err_unknown);
	  n_error_counter++;
	}
	if (!(complete_check & 1)) {
	  n_error_counter++;
	  LOG_PRINT(LOG_WARNING, conv_missing_title);
	}
	if (!(complete_check & 2)) {
	  n_error_counter++;
	  LOG_PRINT(LOG_WARNING, conv_missing_pubyear);
	}
	if (!(complete_check & 4)) {
	  n_error_counter++;
	  LOG_PRINT(LOG_WARNING, conv_missing_journal);
	}
	if (!(complete_check & 8)) {
	  n_error_counter++;
	  LOG_PRINT(LOG_WARNING, conv_missing_volume);
	}
	if (!(complete_check & 16)) {
	  n_error_counter++;
	  LOG_PRINT(LOG_WARNING, conv_missing_page);
	}
	if (!(complete_check & 32)) {
	  n_error_counter++;
	  LOG_PRINT(LOG_WARNING, conv_missing_author);
	}
	if (!(complete_check & 64)) {
	  n_warning_counter++;
	  LOG_PRINT(LOG_WARNING, conv_missing_type);
	}
				
	if (complete_check == 63) {
	  n_error_counter++;
	  LOG_PRINT(LOG_WARNING, conv_type_added);
	}
      }
			
      if (!cstate.n_convert_error && (complete_check == 127 || complete_check == 63)) { /* if all ok, except maybe missing type, dump to output file */
	if (n_waitfortype) {
	  /* start output reference with a blank line */
	  if (fwrite("\nTY  - JOUR\n", sizeof(char), 12, fp_outfile) < 12) {
	    LOG_PRINT(LOG_ERR, "could not write output");
	    cstate.n_outerror = 1;
	  }
	}

	if (*outbuffer) {
	  if (fwrite(outbuffer, sizeof(char), strlen(outbuffer), fp_outfile) < strlen(outbuffer)) {
	    LOG_PRINT(LOG_ERR, "could not write output");
	    cstate.n_outerror = 1;
	  }
	}
      }

    }			
    if (fwrite("ER  - \n", sizeof(char), 7, fp_outfile) < 7) {
      LOG_PRINT(LOG_ERR, "could not write output");
      cstate.n_outerror = 1;
    }

    if (cstate.n_convert_error) {
      sprintf(message_buffer, "%d errors detected in reference %d", cstate.n_convert_error, n_ref_counter);
      LOG_PRINT(LOG_WARNING, message_buffer);
    }

    /* re-initialize for next reference */
    outbuffer[0] = '\0'; 	/* clear output buffer */
    n_waitfortype = 1;
    cstate.n_convert_error = 0;
    complete_check = 0;
    convert_check = 0;
    
    if (cstate.n_outerror == 1 || cstate.n_outomem == 1) {
      LOG_PRINT(LOG_ERR, "could not write output");
    }
  } /* end loop - references main loop */

  sprintf(message_buffer, "%d references converted", n_ref_counter);
  LOG_PRINT(LOG_INFO, message_buffer);

  /* clean up */
  return 1;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  convert_line(): converts a line to RIS format

  int convert_line 1 if successful, 0 if error 

  char *inbuffer a buffer for the incoming data

  int n_source_type type of Medline source

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int convert_line(char* inbuffer, int n_source_type) {
  int n_result;
	
  switch (n_source_type){
  case BIOMEDNET:
    n_result = convert_line_rm(inbuffer);
    break;
  case CCOD:
    n_result = convert_line_kf(inbuffer);
    break;
  case GRATEMED:
    n_result = convert_line_pm(inbuffer);
    break;
  case HEALTHGATE:
    n_result = convert_line_hg(inbuffer);
    break;
  case KNOWFINDER:
    n_result = convert_line_kf(inbuffer);
    break;
  case PUBMED:
    n_result = convert_line_pm(inbuffer);
    break;
  default:
    n_result = 0;
    break;
  }
  return n_result;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  convert_line_kf(): converts a line from Knowledge Finder to RIS format

  int convert_line_kf 1 if successful, 0 if error 

  char *inbuffer a buffer for the incoming data

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int convert_line_kf(char* inbuffer) {
  int retvalue = 0;
  int success = 1;
  char *token;
  char *token_py;
  char *token1;
  char *token2;
  char *end_token;
  char delimiter[2] = ";";
  int i;
	
  /* title or comment on title */
  if (strncmp("TI  - ", inbuffer, 6) == 0) {
    if (complete_check&1) { /* must be a comment or something */
      inbuffer[0] = 'N';
      inbuffer[1] = '1';
      if (!write_out(inbuffer)) {
	return 0; /* error */
      }
      if (!write_out("\n")) {
	return 0; /* error */
      }
    }
    else { /* this is the real title */
      if (!write_out(inbuffer)) {
	return 0; /* error */
      }
      if (!write_out("\n")) {
	return 0; /* error */
      }
      complete_check |= 1;
    }
    retvalue = 1;
  }
	
  /* if it's an abstract, take as is */
  else if (strncmp("AB  - ", inbuffer, 6) == 0) {
    inbuffer[0] = 'N'; /* import into abstract field instead of notes */
    inbuffer[1] = '2';
    if (!write_out(inbuffer)) {
      return 0; /* error */
    }
    retvalue = 1;
  }
	
  /* if it's the source: split it up */
  else if (strncmp("SO  - ", inbuffer, 6) == 0) {
    if (!write_source(inbuffer)) {
      return 0; /* error */
    }
    retvalue = 1;
  }

// if it's the authors, split them up <g>
  else if (strncmp("AU  - ", inbuffer, 6) == 0) {
    token = strtok(&inbuffer[5], delimiter);	// find first token after tag
    if (token == NULL) {
      convert_check |= 4;
      retvalue = 0;
    }
    else {
      while (token != NULL) {
	complete_check |= 32;
	success &= convert_author(&token[1]); // success is FALSE if any conversion fails
	// correct token pointer for leading space
	token = strtok(NULL, delimiter);        // find more tokens until finished
      }
      retvalue = success;
      if (!success) {
	convert_check |= 4;
      }
    }
  }
	
  // if it's the subject headings, convert them to keywords
  else if (strncmp("MJ  - ", inbuffer, 6) == 0 ||
	   strncmp("MN  - ", inbuffer, 6) == 0 ||
	   strncmp("MT  - ", inbuffer, 6) == 0 ||     
	   strncmp("RN  - ", inbuffer, 6) == 0) {       // substances)
    token = strtok(&inbuffer[5], delimiter);   // this preserves the leading space
    while (token != NULL) {
      token1 = strchr(token, '[');
      if (token1 != NULL)	{
	do {
	  *(token1-1) = '\0';
	  token2 = strchr(token1+1, '[');
	  if (token2 != NULL) {
	    *(token2-1) = '\0';
	  }
	  if (!write_out("KW  - ")) {
	    return 0;
	  }
	  if (!write_out(&token[1])) {
	    return 0;
	  }
	  if (!write_out(" ")) {
	    return 0;
	  }
	  if (!write_out(token1)) {
	    return 0;
	  }
	  if (!write_out("\n")) {
	    return 0;
	  }
	  token1 = token2;
	}
	while (token1 != NULL);
      }
      else {
	if (!write_out("KW  - ")) {
	  return 0;
	}
	if (!write_out(&token[1])) {
	  return 0;
	}
	if (!write_out("\n")) {
	  return 0;
	}
      }
      token = strtok(NULL, delimiter);	     // find next token
    }
    retvalue = 1;
  }
	
  // if it's the article type
  else if (strncmp("PT  - ", inbuffer, 6) == 0 || strncmp("RF  - ", inbuffer, 6) == 0) {
    if (strstr(inbuffer, "REVIEW") != NULL || strstr(inbuffer, "Review") != NULL) { // if journal type indicates review
      if (!write_out("KW  - REVIEW\n")) {
	return 0;
      }
    }
    if (strstr(inbuffer, "TUTORIAL") != NULL) {
      if (!write_out("KW  - TUTORIAL\n")) {
	return 0;
      }
    }
    if (strstr(inbuffer, "TECHNICAL REPORT") != NULL) {
      if (!write_out("KW  - TECHNICAL REPORT\n")) {
	return 0;
      }
    }
    if (strstr(inbuffer, "CLINICAL TRIAL") != NULL)	{
      if (!write_out("KW  - CLINICAL TRIAL\n")) {
	return 0;
      }
    }
    complete_check |= 64;
    retvalue = -1;
  }
	
  // if its the comment source
  else if (strncmp("CM  - ", inbuffer, 6) == 0) {
    inbuffer[0] = 'N';                    // import into notes
    inbuffer[1] = '1';
    if (!write_out(inbuffer)) {
      return 0;
    }
    if (!write_out("\n")) {
      return 0;
    }
    retvalue = 1;
  }

  // if its the address field
  else if (strncmp("AD  - ", inbuffer, 6) == 0) {
    if (!write_out(inbuffer)) {
      return 0;
    }
    if (!write_out("\n")) {
      return 0;
    }
    retvalue = 1;
  }

  // if it's a field we're not interested in, simply drop it
  else if (strncmp("MT  - ", inbuffer, 6) == 0 ||
	   strncmp("EM  - ", inbuffer, 6) == 0 ||
	   strncmp("IS  - ", inbuffer, 6) == 0 ||
	   strncmp("LA  - ", inbuffer, 6) == 0 ||
	   strncmp("UI  - ", inbuffer, 6) == 0 ||
	   strncmp("ID  - ", inbuffer, 6) == 0 ||
	   strncmp("TT  - ", inbuffer, 6) == 0 ||
	   strncmp("SI  - ", inbuffer, 6) == 0 ||
	   strncmp("MC  - ", inbuffer, 6) == 0 ||
	   strncmp("PS  - ", inbuffer, 6) == 0 ||
	   strncmp("DA  - ", inbuffer, 6) == 0 ||
	   strncmp("DP  - ", inbuffer, 6) == 0 ||
	   strncmp("IS  - ", inbuffer, 6) == 0 ||
	   strncmp("TA  - ", inbuffer, 6) == 0 ||
	   strncmp("PG  - ", inbuffer, 6) == 0 ||
	   strncmp("SB  - ", inbuffer, 6) == 0 ||
	   strncmp("CY  - ", inbuffer, 6) == 0 ||
	   strncmp("JC  - ", inbuffer, 6) == 0 ||
	   strncmp("AA  - ", inbuffer, 6) == 0 ||
	   strncmp("EM  - ", inbuffer, 6) == 0 ||
	   strncmp("IP  - ", inbuffer, 6) == 0 ||
	   strncmp("VI  - ", inbuffer, 6) == 0 ||
	   strncmp("RO  - ", inbuffer, 6) == 0 ||
	   strncmp("EA  - ", inbuffer, 6) == 0 ||
	   strncmp("CU  - ", inbuffer, 6) == 0 ||
	   strncmp("RF  - ", inbuffer, 6) == 0 ||
	   strncmp("PMID- ", inbuffer, 6) == 0) {
    retvalue = 1;
  }
	
  // if strange unknown tags appear
  else if (!n_ignore_tag && strlen(inbuffer) > 6) {
    if (strncmp("  - ", &inbuffer[2], 4) == 0) {
      retvalue = 0;  // warn user about strange things in file
      convert_check |= 8;
    }
  }
  else {
    retvalue = 1;   
  }
  return retvalue;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  convert_line_pm(): converts a line from PubMed to RIS format

  int convert_line_pm 1 if successful, 0 if error 

  char *inbuffer a buffer for the incoming data

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int convert_line_pm(char* inbuffer) {
  int retvalue = 0;
  int success = 1;
  char *token, *token_py, *token1, *token2;
  char delimiter[2] = "/";
  int i;
	
  /* title or comment on title */
  if (strncmp("TI  - ", inbuffer, 6) == 0) {
    if (complete_check&1) { /* must be a comment or something */
      inbuffer[0] = 'N';
      inbuffer[1] = '1';
      if (!write_out(inbuffer)) {
	return 0; /* error */
      }
      if (!write_out("\n")) {
	return 0; /* error */
      }
    }
    else { /* this is the real title */
      if (!write_out(inbuffer)) {
	return 0; /* error */
      }
      if (!write_out("\n")) {
	return 0; /* error */
      }
      complete_check |= 1;
    }
    retvalue = 1;
  }
	
  // if it's an abstract, take as is
  else if (strncmp("AB  - ", inbuffer, 6) == 0) {
    inbuffer[0] = 'N';                    // import into abstract field instead of notes
    inbuffer[1] = '2';
    if (!write_out(inbuffer)) {
      return 0; /* error */
    }
    if (!write_out("\n")) {
      return 0; /* error */
    }
    retvalue = 1;
  }
	
  /* if it's the source: split it up */
  else if (strncmp("SO  - ", inbuffer, 6) == 0) {
    if (!write_source(inbuffer)) {
      return 0; /* error */
    }
    retvalue = 1;
  }

  // if it's the authors, split them up <g>
  else if (strncmp("AU  - ", inbuffer, 6) == 0) {
    token = strtok(&inbuffer[5], delimiter);	// find first token after tag
    if (token == NULL) {
      convert_check |= 4;
      retvalue = 0;
    }
    else {
      while (token != NULL) {
	complete_check |= 32;
	success &= convert_author(&token[1]); // success is FALSE if any conversion fails
	// correct token pointer for leading space
	token = strtok(NULL, delimiter);        // find more tokens until finished
      }
      retvalue = success;
      if (!success) {
	convert_check |= 4;
      }
    }
  }
	
  // if it's the subject headings, convert them to keywords
  else if (strncmp("MH  - ", inbuffer, 6) == 0 ||     
	   strncmp("RN  - ", inbuffer, 6) == 0) {      // substance name
    do {                                         // remove asterisks
      token1 = strchr(inbuffer, '*');
      if (token1 != NULL) {
	*token1 = '\0';
	strcat(inbuffer, token1+1);
      }
    } while (token1 != NULL);
		
    token1 = strchr(&inbuffer[6], '/');
    if (token1 != NULL) {
      *token1 = '\0';
      token1++;					// adjust to start of subheadings
      token = strtok(token1, delimiter);   
      while (token != NULL) {
	if (!write_out("KW  - ")) {
	  return 0; /* error */
	}
	if (!write_out(&inbuffer[6])) {
	  return 0; /* error */
	}
	if (!write_out(" [")) {
	  return 0; /* error */
	}
	if (!write_out(token)) {
	  return 0; /* error */
	}
	if (!write_out("]\n")) {
	  return 0; /* error */
	}
	token = strtok(NULL, delimiter);	     // find next token
      }
    }
    else {
      if (!write_out("KW  - ")) {
	return 0; /* error */
      }
      if (!write_out(&inbuffer[6])) {
	return 0; /* error */
      }
      if (!write_out("\n")) {
	return 0; /* error */
      }
    }
    retvalue = 1;
  }
	
  // if it's the article type
  else if (strncmp("PT  - ", inbuffer, 6) == 0 || strncmp("RF  - ", inbuffer, 6) == 0) {
    if (strstr(inbuffer, "REVIEW") != NULL || strstr(inbuffer, "Review") != NULL) { // if journal type indicates review
      if (!write_out("KW  - REVIEW\n")) {
	return 0; /* error */
      }
    }
    if (strstr(inbuffer, "TUTORIAL") != NULL) {
      if (!write_out("KW  - TUTORIAL\n")) {
	return 0; /* error */
      }
    }
    if (strstr(inbuffer, "TECHNICAL REPORT") != NULL) {
      if (!write_out("KW  - TECHNICAL REPORT\n")) {
	return 0; /* error */
      }
    }
    if (strstr(inbuffer, "CLINICAL TRIAL") != NULL) {
      if (!write_out("KW  - CLINICAL TRIAL\n")) {
	return 0; /* error */
      }
    }
    //strcpy(inbuffer, "TY  - JOUR\r\n");    // to make sure all info is imported by RM, ALL articles are JOUR
    complete_check |= 64;
    retvalue = -1;
  }
	
  // if its the comment source
  else if (strncmp("CM  - ", inbuffer, 6) == 0) {
    inbuffer[0] = 'N';                    // import into notes
    inbuffer[1] = '1';
    if (!write_out(inbuffer)) {
      return 0; /* error */
    }
    if (!write_out("\n")) {
      return 0; /* error */
    }
    retvalue = 1;
  }

  // if its the address field
  else if (strncmp("AD  - ", inbuffer, 6) == 0) {
    if (!write_out(inbuffer)) {
      return 0; /* error */
    }
    if (!write_out("\n")) {
      return 0; /* error */
    }
    retvalue = 1;
  }

  /* a web url */
  /* "URL  - " is the old form. Newer datasets use "URLF- " for full text
     and "URLS- " for the abstract link. We use only the link to the full
     text here as the abstract is readily available in Medline anyway */
  else if (strncmp("URL - ", inbuffer, 6) == 0 || strncmp ("URLF- ", inbuffer, 6) == 0) {
    inbuffer[2] = ' '; /* convert to 'UR  - ' */
    inbuffer[3] = ' '; 
    if (!write_out(inbuffer)) {
      return 0; /* error */
    }
    if (!write_out("\n")) {
      return 0; /* error */
    }
    retvalue = 1;
  }

  // if it's a field we're not interested in, simply drop it
  else if (strncmp("MT  - ", inbuffer, 6) == 0 ||
	   strncmp("EM  - ", inbuffer, 6) == 0 ||
	   strncmp("IS  - ", inbuffer, 6) == 0 ||
	   strncmp("LA  - ", inbuffer, 6) == 0 ||
	   strncmp("UI  - ", inbuffer, 6) == 0 ||
	   strncmp("ID  - ", inbuffer, 6) == 0 ||
	   strncmp("TT  - ", inbuffer, 6) == 0 ||
	   strncmp("SI  - ", inbuffer, 6) == 0 ||
	   strncmp("MC  - ", inbuffer, 6) == 0 ||
	   strncmp("PS  - ", inbuffer, 6) == 0 ||
	   strncmp("DA  - ", inbuffer, 6) == 0 ||
	   strncmp("DP  - ", inbuffer, 6) == 0 ||
	   strncmp("IS  - ", inbuffer, 6) == 0 ||
	   strncmp("TA  - ", inbuffer, 6) == 0 ||
	   strncmp("PG  - ", inbuffer, 6) == 0 ||
	   strncmp("SB  - ", inbuffer, 6) == 0 ||
	   strncmp("CY  - ", inbuffer, 6) == 0 ||
	   strncmp("JC  - ", inbuffer, 6) == 0 ||
	   strncmp("AA  - ", inbuffer, 6) == 0 ||
	   strncmp("EM  - ", inbuffer, 6) == 0 ||
	   strncmp("IP  - ", inbuffer, 6) == 0 ||
	   strncmp("VI  - ", inbuffer, 6) == 0 ||
	   strncmp("RO  - ", inbuffer, 6) == 0 ||
	   strncmp("EA  - ", inbuffer, 6) == 0 ||
	   strncmp("CU  - ", inbuffer, 6) == 0 ||
	   strncmp("RF  - ", inbuffer, 6) == 0 ||
	   strncmp("PMID- ", inbuffer, 6) == 0) {
    retvalue = 1;
  }
	
  // if strange unknown tags appear
  else if (!n_ignore_tag && strlen(inbuffer) > 6) {
    if (strncmp("  - ", &inbuffer[2], 4) == 0) {
      retvalue = 0;  // warn user about strange things in file
      convert_check |= 8;
    }
  }
  else {
    retvalue = 1;   
  }
  return retvalue;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  convert_line_rm(): converts a line from near-RIS to RIS format

  int convert_line_rm 1 if successful, 0 if error 

  char *inbuffer a buffer for the incoming data

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int convert_line_rm(char* inbuffer) {
  int retvalue = 1;
  int success = 1;
  char *token;
  char delimiter[2] = ";";
	
  /* title or comment on title */
  if (strncmp("TI  - ", inbuffer, 6) == 0) {
    if (complete_check&1) { /* must be a comment or something */
      inbuffer[0] = 'N';
      inbuffer[1] = '1';
      if (!write_out(inbuffer)) {
	return 0; /* error */
      }
      if (!write_out("\n")) {
	return 0; /* error */
      }
    }
    else { /* this is the real title */
      if (!write_out(inbuffer)) {
	return 0; /* error */
      }
      if (!write_out("\n")) {
	return 0; /* error */
      }
      complete_check |= 1;
    }
    retvalue = 1;
  }
	
  // if it's an abstract, take as is
  else if (strncmp("AB  - ", inbuffer, 6) == 0) {
    if (strncmp("Address:", inbuffer+6, 8) == 0) {
      inbuffer[0] = 'A';                    // import into notes
      inbuffer[1] = 'D';
    }
    else if (strncmp("Type:", inbuffer+6, 5) == 0) {
      inbuffer[0] = 'N';                    // import into notes
      inbuffer[1] = '1';
    }
    else if (strncmp("Language:", inbuffer+6, 9) == 0) {
      inbuffer[0] = 'N';                    // import into notes
      inbuffer[1] = '1';
    }
    else {   // its the abstract proper
      inbuffer[0] = 'N';                    // import into abstract field instead of notes
      inbuffer[1] = '2';
    }
    if (!write_out(inbuffer)) {
      return 0; /* error */
    }
    if (!write_out("\n")) {
      return 0; /* error */
    }
    retvalue = 1;
  }
	
  else if (strncmp("N1  - ", inbuffer, 6) == 0) {
    if (!write_out(inbuffer)) {
      return 0; /* error */
    }
    if (!write_out("\n")) {
      return 0; /* error */
    }
    retvalue = 1;
  }
	
  else if (strncmp("PY  - ", inbuffer, 6) == 0) {
    if (!write_out(inbuffer)) {
      return 0; /* error */
    }
    if (!write_out("\n")) {
      return 0; /* error */
    }
    complete_check |= 2;
    retvalue = 1;
  }
	
  else if (strncmp("JO  - ", inbuffer, 6) == 0) {
    if (!write_out(inbuffer)) {
      return 0; /* error */
    }
    if (!write_out("\n")) {
      return 0; /* error */
    }
    complete_check |= 4;
    retvalue = 1;
  }

  else if (strncmp("VL  - ", inbuffer, 6) == 0) {
    if (!write_out(inbuffer)) {
      return 0; /* error */
    }
    if (!write_out("\n")) {
      return 0; /* error */
    }
    complete_check |= 8;
    retvalue = 1;
  }

  else if (strncmp("IS  - ", inbuffer, 6) == 0) {
    if (!write_out(inbuffer)) {
      return 0; /* error */
    }
    if (!write_out("\n")) {
      return 0; /* error */
    }
    retvalue = 1;
  }

  else if (strncmp("SP  - ", inbuffer, 6) == 0) {
    if (!write_out(inbuffer)) {
      return 0; /* error */
    }
    if (!write_out("\n")) {
      return 0; /* error */
    }
    complete_check |= 16;
    retvalue = 1;
  }

  else if (strncmp("EP  - ", inbuffer, 6) == 0) {
    if (!write_out(inbuffer)) {
      return 0; /* error */
    }
    if (!write_out("\n")) {
      return 0; /* error */
    }
    retvalue = 1;
  }

  else if (strncmp("AU  - ", inbuffer, 6) == 0) {
    if (!write_out(inbuffer)) {
      return 0; /* error */
    }
    if (!write_out("\n")) {
      return 0; /* error */
    }
    complete_check |= 32;
    retvalue = 1;
  }

  else if (strncmp("KW  - ", inbuffer, 6) == 0) {
    if (!write_out(inbuffer)) {
      return 0; /* error */
    }
    if (!write_out("\n")) {
      return 0; /* error */
    }
    retvalue = 1;
  }

  // if it's the article type
  else if (strncmp("TY  - ", inbuffer, 6) == 0) {
    complete_check |= 64;
    retvalue = -1;
  }
	
  // if it's the end of ref tag
  else if (strncmp("ER  -", inbuffer, 5) == 0) {
    retvalue = 1;
  }
	
  // if strange unknown tags appear
  else if (!n_ignore_tag && strlen(inbuffer) > 6)	{
    if (strncmp("  - ", &inbuffer[2], 4) == 0) {
      retvalue = 0;  // warn user about strange things in file
      convert_check |= 8;
    }
  }
  else {
    retvalue = 1;   
  }
  return retvalue;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  convert_line_hg(): converts a line from Healthgate to RIS format

  int convert_line_hg 1 if successful, 0 if error 

  char *inbuffer a buffer for the incoming data

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int convert_line_hg(char* inbuffer) {
  int retvalue = 0;
  int success = 1;
  char *token, *token_py, *token1, *token2;
  char delimiter[2] = ";";
  int i;
	
  /* title or comment on title */
  if (strncmp("TI  - ", inbuffer, 6) == 0) {
    if (complete_check&1) { /* must be a comment or something */
      inbuffer[0] = 'N';
      inbuffer[1] = '1';
      if (!write_out(inbuffer)) {
	return 0; /* error */
      }
      if (!write_out("\n")) {
	return 0; /* error */
      }
    }
    else { /* this is the real title */
      if (!write_out(inbuffer)) {
	return 0; /* error */
      }
      if (!write_out("\n")) {
	return 0; /* error */
      }
      complete_check |= 1;
    }
    retvalue = 1;
  }
	
  // if it's an abstract, take as is
  else if (strncmp("AB  - ", inbuffer, 6) == 0) {
    inbuffer[0] = 'N';                    // import into abstract field instead of notes
    inbuffer[1] = '2';
    if (!write_out(inbuffer)) {
      return 0; /* error */
    }
    if (!write_out("\n")) {
      return 0; /* error */
    }
    retvalue = 1;
  }
	
  /* if it's the source: split it up */
  else if (strncmp("SO  - ", inbuffer, 6) == 0) {
    if (!write_source(inbuffer)) {
      return 0; /* error */
    }
    retvalue = 1;
  }

  // if it's the authors, split them up <g>
  else if (strncmp("AU  - ", inbuffer, 6) == 0) {
    token = strtok(&inbuffer[5], delimiter);	// find first token after tag
    if (token == NULL) {
      convert_check |= 4;
      retvalue = 0;
    }
    else {
      while (token != NULL) {
	complete_check |= 32;
	success &= convert_author(&token[1]); // success is FALSE if any conversion fails
	// correct token pointer for leading space
	token = strtok(NULL, delimiter);        // find more tokens until finished
      }
      retvalue = success;
      if (!success) {
	convert_check |= 4;
      }
    }
  }
	
  // if it's the subject headings, convert them to keywords
  else if (strncmp("MJ  - ", inbuffer, 6) == 0 ||
	   strncmp("MN  - ", inbuffer, 6) == 0 ||
	   strncmp("MH  - ", inbuffer, 6) == 0 ||     
	   strncmp("RN  - ", inbuffer, 6) == 0) {      // substances
    do {                                         // remove asterisks
      token1 = strchr(inbuffer, '*');
      if (token1 != NULL) {
	*token1 = '\0';
	strcat(inbuffer, token1+1);
      }
    } while (token1 != NULL);
		
    token1 = strchr(inbuffer, '|');    // this removes keyword subclassification
    if (token1 != NULL) {            // TODO: convert subclassifications into plain text
      *token1 = '\0';
    }
			
    if (token1 != NULL) {					// if there are subheadings
      token1++;							// promote to start of subheadings
      if (strlen(token1) != 0) {
	token = strtok(token1, "/");
				
	while (token != NULL) {
	  if (!write_out("KW  - ")) {
	    return 0; /* error */
	  }
	  if (!write_out(&inbuffer[6])) {
	    return 0; /* error */
	  }
	  add_subheading(token);
	  if (!write_out("\n")) {
	    return 0; /* error */
	  }
	  token = strtok(NULL, "/");
	}
      }
      else {           // sometimes "|*" is the sole subheading, this is removed
	if (!write_out("KW  - ")) {
	  return 0; /* error */
	}
	if (!write_out(&inbuffer[6])) {
	  return 0; /* error */
	}
	if (!write_out("\n")) {
	  return 0; /* error */
	}
      }
    }
    else {
      if (!write_out("KW  - ")) {
	return 0; /* error */
      }
      if (!write_out(&inbuffer[6])) {
	return 0; /* error */
      }
      if (!write_out("\n")) {
	return 0; /* error */
      }
    }
    retvalue = 1;
  }
	
  // if it's the article type
  else if (strncmp("PT  - ", inbuffer, 6) == 0) {
    if (strstr(inbuffer, "REVIEW") != NULL) {     // if journal type indicates review
      if (!write_out("KW  - REVIEW\n")) {
	return 0; /* error */
      }
    }
    if (strstr(inbuffer, "TUTORIAL") != NULL) {
      if (!write_out("KW  - TUTORIAL\n")) {
	return 0; /* error */
      }
    }
    if (strstr(inbuffer, "TECHNICAL REPORT") != NULL) {
      if (!write_out("KW  - TECHNICAL REPORT\n")) {
	return 0; /* error */
      }
    }
    if (strstr(inbuffer, "CLINICAL TRIAL") != NULL) {
      if (!write_out("KW  - CLINICAL TRIAL\n")) {
	return 0; /* error */
      }
    }
    //strcpy(inbuffer, "TY  - JOUR\r\n");    // to make sure all info is imported by RM, ALL articles are JOUR
    complete_check |= 64;
    retvalue = -1;
  }
	
  // if its the comment source
  else if (strncmp("CM  - ", inbuffer, 6) == 0) {
    inbuffer[0] = 'N';                    // import into notes
    inbuffer[1] = '1';
    if (!write_out(inbuffer)) {
      return 0; /* error */
    }
    if (!write_out("\n")) {
      return 0; /* error */
    }
    retvalue = 1;
  }

  // if its the address field
  else if (strncmp("AD  - ", inbuffer, 6) == 0) {
    if (!write_out(inbuffer)) {
      return 0; /* error */
    }
    if (!write_out("\n")) {
      return 0; /* error */
    }
    retvalue = 1;
  }

  // if it's a field we're not interested in, simply drop it
  else if (strncmp("MT  - ", inbuffer, 6) == 0 ||
	   strncmp("EM  - ", inbuffer, 6) == 0 ||
	   strncmp("IS  - ", inbuffer, 6) == 0 ||
	   strncmp("LA  - ", inbuffer, 6) == 0 ||
	   strncmp("UI  - ", inbuffer, 6) == 0 ||
	   strncmp("ID  - ", inbuffer, 6) == 0 ||
	   strncmp("TT  - ", inbuffer, 6) == 0 ||
	   strncmp("SI  - ", inbuffer, 6) == 0 ||
	   strncmp("MC  - ", inbuffer, 6) == 0 ||
	   strncmp("PS  - ", inbuffer, 6) == 0 ||
	   strncmp("DA  - ", inbuffer, 6) == 0 ||
	   strncmp("DP  - ", inbuffer, 6) == 0 ||
	   strncmp("IS  - ", inbuffer, 6) == 0 ||
	   strncmp("TA  - ", inbuffer, 6) == 0 ||
	   strncmp("PG  - ", inbuffer, 6) == 0 ||
	   strncmp("SB  - ", inbuffer, 6) == 0 ||
	   strncmp("CY  - ", inbuffer, 6) == 0 ||
	   strncmp("JC  - ", inbuffer, 6) == 0 ||
	   strncmp("AA  - ", inbuffer, 6) == 0 ||
	   strncmp("EM  - ", inbuffer, 6) == 0 ||
	   strncmp("IP  - ", inbuffer, 6) == 0 ||
	   strncmp("VI  - ", inbuffer, 6) == 0 ||
	   strncmp("RO  - ", inbuffer, 6) == 0 ||
	   strncmp("EA  - ", inbuffer, 6) == 0 ||
	   strncmp("CU  - ", inbuffer, 6) == 0 ||
	   strncmp("RF  - ", inbuffer, 6) == 0 ||
	   strncmp("PMID- ", inbuffer, 6) == 0) {
    retvalue = 1;
  }
	
  // if strange unknown tags appear
  else if (!n_ignore_tag && strlen(inbuffer) > 6) {
    if (strncmp("  - ", &inbuffer[2], 4) == 0) {
      retvalue = 0;  // warn user about strange things in file
      convert_check |= 8;
    }
  }
  else {
    retvalue = 1;   
  }
  return retvalue;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  skip_lines_fp(): skips leading empty lines in a file

  int skip_lines_fp returns 1 if EOF was reached, 0 otherwise

  char* inbuffer buffer for incoming data

  FILE* fp_infile ptr to FILE to read from

  int* n_read_err ptr to read error indicator

  int* n_read_incomplete ptr to read incomplete indicator

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int skip_lines_fp(char* inbuffer, FILE* fp_infile, struct convstate* ptr_cstate) {
  int n_eof = 0;
  int n_inbuf_len; /* length of string in inbuffer */
  char* read_result; /* return value of fgets */

  while (1) { /* exit this loop with break */
    if ((read_result = fgets(inbuffer, INFILE_CHUNK_SIZE, fp_infile)) == NULL) {
      if (errno != EOF) {
	ptr_cstate->n_read_err = 1;
	n_eof = 1;
      }
      else {
	n_eof = 1;
      }
/*  	printf("fgets cant read\n"); */
	/*  return 1; */ /* doesn't matter whether EOF or read error */
      break;
    }
    if (*read_result != '\n' && *read_result != '\r') {
      n_inbuf_len = strlen(inbuffer);
      if (n_inbuf_len == INFILE_CHUNK_SIZE-1 && 
	  inbuffer[n_inbuf_len-2] != '\n') { /* fgets read an incomplete string (not finished w/ \n) */
	ptr_cstate->n_read_incomplete = 1;
      }
      else {
	stripwhite(inbuffer, 2, 1); /* remove lf */
      }
      break;
    }
  } /* end while loop - skipping empty lines */
  return n_eof;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  skip_lines_s(): skips leading empty lines in a string

  int skip_lines_s returns 1 if EOS was reached, 0 otherwise

  char* inbuffer buffer for incoming data

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int skip_lines_s(char* inbuffer) {
  while (*inbuffer < 32 && *inbuffer != '\0') {
    inbuffer++;
  }
  if (!*inbuffer) {
    return 1; /* end of string */
  }
  else {
    return 0;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  convert_ref_fp(): reads data and converts until reference complete

  int convert_ref_fp returns 1 if EOS was reached, 0 otherwise

  char** inbuffer ptr to ptr to buffer for incoming data

  int* n_waitfortype ptr to flag, will be set if type field is found

  int n_source_type type of data source

  struct convstate* ptr_cstate ptr to structure with error flags

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int convert_ref_fp(char** inbuffer, int* n_waitfortype, int n_source_type, struct convstate* ptr_cstate) {
  char* read_result; /* return value of fgets */
  char* sup_inbuffer; /* supplementary inbuffer to read multi-lines */
  char* strip_inbuffer; /* used to strip off whitespace */
  char* new_inbuffer; /* used to reallocate inbuffer */
  int n_eof = 0;
  int n_convline_result = 0; /* return value of convert_line() */
  int nextchar; /* used to test multi-line entries */

  /* get us some memory for the input */
  if ((sup_inbuffer = (char*)malloc(INFILE_CHUNK_SIZE)) == NULL) {
    LOG_PRINT(LOG_CRIT, "out of memory");
    return -1;
  }

  do { /* loop until next empty line indicating end of reference */
    n_convline_result = convert_line(*inbuffer, n_source_type);
    if (n_convline_result == -1)	{ /* this was the type field */
      *n_waitfortype = 0;
      if (*outbuffer) {
	if (fwrite(outbuffer, sizeof(char), strlen(outbuffer), fp_outfile) < strlen(outbuffer)) {
	  LOG_PRINT(LOG_WARNING, "writing output failed");
	  ptr_cstate->n_outerror = 1;
	}
	outbuffer[0] = '\0';
      }
    }
    else if (n_convline_result == 0) {  /* converting error */
      ptr_cstate->n_convert_error = 1;
    }
    else if (n_convline_result == -2) {    /* malloc failed */
      ptr_cstate->n_outomem = 1;
    }
    /* else was ok */
    if (ptr_cstate->n_read_incomplete) {
      while (ptr_cstate->n_read_incomplete) {
	ptr_cstate->n_read_incomplete = 0;
	if ((read_result = fgets(*inbuffer, INFILE_CHUNK_SIZE, fp_infile)) == NULL) {
	  if (errno != EOF) {
	    ptr_cstate->n_read_err = 1;
	    n_eof = 1;
	  }
	  else {
	    n_eof = 1;
	  }
	}
	if (*read_result != '\n' && *read_result != '\r') {
	  n_inbuffer_len = strlen(*inbuffer);
	  if (n_inbuffer_len == INFILE_CHUNK_SIZE-1 && 
	      (*inbuffer)[n_inbuffer_len-2] != '\n') { /* fgets read an incomplete string (not finished w/ \n) */
	    ptr_cstate->n_read_incomplete = 1;
	    if (fwrite(*inbuffer, sizeof(char), strlen(*inbuffer), fp_outfile) < strlen(outbuffer)) {
	      LOG_PRINT(LOG_WARNING, "writing output failed");
	      ptr_cstate->n_outerror = 1;
	    }
	  }
	  else {
	    stripwhite(*inbuffer, 2, 1); /* remove lf */
	  }
	} /* else: line is completed right after previous chunk, ignore */
      } /* end while (n_read_incomplete) */
    }
    else {
      if ((read_result = fgets(*inbuffer, INFILE_CHUNK_SIZE, fp_infile)) == NULL) {
	if (errno != EOF) {
	  ptr_cstate->n_read_err = 1;
	  n_eof = 1;
	}
	else {
	  n_eof = 1;
	}
	break; /* doesn't matter whether EOF or read error */
      }
    }
    
    n_inbuffer_len = strlen(*inbuffer);
    if (n_inbuffer_len == INFILE_CHUNK_SIZE-1 && 
	*inbuffer[n_inbuffer_len-2] != '\n') { /* fgets read an incomplete string (not finished w/ \n) */
      ptr_cstate->n_read_incomplete = 1;
    }
    else if (**inbuffer != '\n' && **inbuffer != '\r') {
      stripwhite(*inbuffer, 2, 1); /* remove lf */
      /* now that we have a line, we must check whether the next line
	 also starts with a tag. Otherwise the next line is a continuation 
	 of the current line */
      while (1) { /* exit this loop with break */
	nextchar = getc(fp_infile); /* peek ahead */
	ungetc(nextchar, fp_infile); /* and put the char back in line */

	if (nextchar == (int)' ' || nextchar == (int)'\t') { /* this is a continued line */
	  if ((read_result = fgets(sup_inbuffer, INFILE_CHUNK_SIZE, fp_infile)) == NULL) {
	    if (errno != EOF) {
	      ptr_cstate->n_read_err = 1;
	    }
	    else {
	      n_eof = 1;
	    }
	    break; /* doesn't matter whether EOF or read error */
	  }
	  if (*read_result != '\n' && *read_result != '\r') {
	    strip_inbuffer = stripwhite(sup_inbuffer, 0, 1);

	    if ((new_inbuffer = mstrcat(*inbuffer, " ", &n_inbuffer_len, 0)) == NULL) {
	      ptr_cstate->n_outomem = 1;
	      LOG_PRINT(LOG_WARNING, "out of memory");
	      break;
	    }
	    else {
	      *inbuffer = new_inbuffer;
	    }
	    if ((new_inbuffer = mstrcat(*inbuffer, strip_inbuffer, &n_inbuffer_len, 0)) == NULL) {
	      ptr_cstate->n_outomem = 1;
	      LOG_PRINT(LOG_WARNING, "out of memory");
	      break;
	    }
	    else {
	      *inbuffer = new_inbuffer;
	    }
	  }
	  /* else found an empty line = end of reference */
	}
	else {
	  break; /* found next full line */
	}
      }
    }
    /* else we have an empty line = end of reference */

  }while (*read_result != '\n' && *read_result != '\r');

  free(sup_inbuffer);
  return n_eof;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  convert_ref_s(): reads data from string and converts until reference
                   complete

  int convert_ref_s returns 1 if EOS was reached, 0 otherwise

  char** inbuffer ptr to ptr to buffer for incoming data

  int* n_waitfortype ptr to flag, will be set if type field is found

  int n_source_type type of data source

  struct convstate* ptr_cstate ptr to structure with error flags

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
int convert_ref_s(char** inbuffer, int* n_waitfortype, int n_source_type, struct convstate* ptr_cstate) {
  char* this_line;
  char* next_line;
  int n_convline_result = 0; /* return value of convert_line() */

  next_line = *inbuffer;

  /* loop until end of string or empty line */
  do {
    this_line = getln_s(&next_line);
    LOG_PRINT(LOG_DEBUG, this_line);
    n_convline_result = convert_line(this_line, n_source_type);
    if (n_convline_result == -1)	{ /* this was the type field */
      *n_waitfortype = 0;
      if (*outbuffer) {
	if (fwrite(outbuffer, sizeof(char), strlen(outbuffer), fp_outfile) < strlen(outbuffer)) {
	  LOG_PRINT(LOG_WARNING, "writing output failed");
	  ptr_cstate->n_outerror = 1;
	}
	outbuffer[0] = '\0';
      }
    }
    else if (n_convline_result == 0) {  /* converting error */
      ptr_cstate->n_convert_error = 1;
    }
    else if (n_convline_result == -2) {    /* malloc failed */
      ptr_cstate->n_outomem = 1;
    }
  } while (*next_line != '\0' && *next_line != '\n' && *next_line != '\r');

  if (*next_line == '\0') {
    return 1; /* end of string */
  }
  else {
    return 0;
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  getln_s(): reads one line from a string

  char getln_s returns ptr to string containing the line

  char** inbuffer ptr to ptr to buffer for incoming data. The ptr
                  will point to the start of the next line (or to \0)
		  when the function returns

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
char* getln_s(char** inbuffer) {
  char* linebreak;
  char* linestart;
  char* stripped_inbuffer;

  /* remove leading whitespace */
  stripped_inbuffer = stripwhite(*inbuffer, 1, 1);

  linebreak = stripped_inbuffer;

  while (*linebreak != '\0') {
    /* if we hit the end of a line, we need to test whether a
       continuation of this line follows */
    if (*linebreak == '\r' || *linebreak == '\n') {
      if (*(linebreak+1) == ' ' || *(linebreak+1) == '\t') {
	*linebreak = ' ';
	/* continuation lines start with whitespace which we
	   remove using memmove */
	linestart = linebreak;
	while (*linestart != '\0'
	       && (*linestart == ' ' || *linestart == '\t')) {
	  linestart++;
	}
	linebreak++;
	memmove(linebreak, linestart, strlen(linestart)+1);
      }
      /* we could also look at cr/lf limited data */
      else if (*linebreak == '\r' && *(linebreak+1) == '\n') {
	if (*(linebreak+2) == ' ' || *(linebreak+2) == '\t') {
	  *linebreak = ' ';
	  *(linebreak+1) = ' ';

	  /* continuation lines start with whitespace which we
	     remove using memmove */
	  linestart = linebreak;
	  while (*linestart != '\0'
		 && (*linestart == ' ' || *linestart == '\t')) {
	    linestart++;
	  }
	  linebreak++;
	  memmove(linebreak, linestart, strlen(linestart)+1);
	}
	else {
	  *linebreak = '\0';
	  linebreak++;
	  break;
	}
      }
      else {
	break;
      }
    }
    else {
      linebreak++;
    }
  }

  if (*linebreak == '\0') { /* end of string */
    *inbuffer = linebreak;
  }
  else {                    /* terminate line and set ptr to next line */
    *linebreak = '\0';
    *inbuffer = linebreak + 1;
  }

  return stripped_inbuffer;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  Log_print(): writes a log message

  void log_print

  int priority the priority level of the log message as in syslog.h

  char* string a string containing the message to be logged

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void log_print(int priority, char* string) {
  /* we must have this fn in the file with main() because FILE* 
     cannot be declared extern (??) */
  time_t the_time;
  char timestring[256];
  FILE* new_fp_log_file;

  if (n_log_dest == 0) { /* output on stderr */
    fprintf(stderr, "%s\n", string);
  }
  else if (n_log_dest == 1) { /* output via syslog */
    syslog(priority, "%s pid=%d", string, getpid());
  }
  else { /* output in user-defined logfile */
    time(&the_time);
    strftime(timestring, 256, "%a %b %d %H:%M:%S %Y", gmtime(&the_time));
    fprintf(fp_logfile, "%d:pid=%d:%s:%s\n", priority, getpid(), timestring, string);
  }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  pipehandler(): handler for the SIGPIPE signal. Sets the global
                 variable n_broken_pipe to non-zero. This condition
                 can be used by a function writing to a pipe to abort
                 further write attempts. Any function that uses
                 n_broken_pipe should reset it to zero when it is done
                 dealing with the broken pipe.

  void pipehandler

  int sig the received signal

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void pipehandler(int sig) {
  n_broken_pipe = 1;
}

#ifdef REFDB_CGI
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  alrmhandler(): handler for the SIGALRM signal. The app will exit.
                 Purpose of this behaviour is to kill hanging CGI
		 programs that could otherwise accumulate until the
		 server dies of memory exhaustion

  void alrmhandler

  int sig the received signal

  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void alrmhandler(int sig) {
    LOG_PRINT(LOG_ERR, "received suicide alarm and exited");
    exit(2);
}
#endif
