<?xml version="1.0" encoding="utf-8"?>
<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
                version="1.0">

  <!-- Use this stylesheet to extract unabbreviated journal title
       words from the list of PubMed Serials available at
       http://www.ncbi.nlm.nih.gov/entrez/citmatch_help.html#JournalLists

       If available, this script evaluates the ISOAbbreviation field,
       otherwise it looks at MedlineTA. In both cases, the items in
       the abbreviation are compared against the full title to
       identify any non-abbreviated words. As this script emits tons
       of duplicates, it is advisable to run the results through sort
       and uniq.
       -->

  <xsl:output method="xml"/>

  <xsl:template match="/">
    <RefDBWordSet>
      <xsl:apply-templates select="SerialsSet/Serial">
        <xsl:sort/>
      </xsl:apply-templates>
    </RefDBWordSet>
  </xsl:template>

  <xsl:template match="Serial">
    <xsl:choose>
      <xsl:when test="not(ISOAbbreviation)">
        <xsl:call-template name="find-med-notabbrev">
          <xsl:with-param name="title" select="Title"/>
          <xsl:with-param name="abbrev" select="MedlineTA"/>
        </xsl:call-template>
      </xsl:when>
      <xsl:otherwise>
        <xsl:call-template name="find-iso-notabbrev">
          <xsl:with-param name="title" select="Title"/>
          <xsl:with-param name="abbrev" select="ISOAbbreviation"/>
        </xsl:call-template>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>

  <xsl:template name="find-med-notabbrev">
    <xsl:param name="title"/>
    <xsl:param name="abbrev"/>
    <xsl:variable name="fulltitle" select="concat($title,' ')"/>
    <xsl:if test="$title and $abbrev">
      <xsl:choose>
        <xsl:when test="contains($abbrev,' ')">
          <xsl:call-template name="find-abbrev-in-title">
            <xsl:with-param name="title" select="$fulltitle"/>
            <xsl:with-param name="word" select="substring-before($abbrev,' ')"/>
          </xsl:call-template>
          <xsl:call-template name="find-med-notabbrev">
            <xsl:with-param name="title" select="$title"/>
            <xsl:with-param name="abbrev" select="substring-after($abbrev,' ')"/>
          </xsl:call-template>
        </xsl:when>
        <xsl:otherwise>
          <xsl:call-template name="find-abbrev-in-title">
            <xsl:with-param name="title" select="$fulltitle"/>
            <xsl:with-param name="word" select="$abbrev"/>
          </xsl:call-template>
        </xsl:otherwise>
      </xsl:choose>
    </xsl:if>
  </xsl:template>

  <xsl:template name="find-iso-notabbrev">
    <xsl:param name="title"/>
    <xsl:param name="abbrev"/>
    <xsl:variable name="fulltitle" select="concat($title,' ')"/>
    <xsl:if test="$title and $abbrev">
      <xsl:choose>
        <xsl:when test="contains($abbrev,'. ')">
          <xsl:call-template name="find-abbrev-in-title">
            <xsl:with-param name="title" select="$fulltitle"/>
            <xsl:with-param name="word" select="substring-before($abbrev,'.')"/>
          </xsl:call-template>
          <xsl:call-template name="find-iso-notabbrev">
            <xsl:with-param name="title" select="$title"/>
            <xsl:with-param name="abbrev" select="substring-after($abbrev,' ')"/>
          </xsl:call-template>
        </xsl:when>
        <xsl:when test="contains($abbrev,'.')">
          <xsl:call-template name="find-abbrev-in-title">
            <xsl:with-param name="title" select="$fulltitle"/>
            <xsl:with-param name="word" select="substring-before($abbrev,'.')"/>
          </xsl:call-template>
        </xsl:when>
        <xsl:otherwise>
          <xsl:call-template name="find-abbrev-in-title">
            <xsl:with-param name="title" select="$fulltitle"/>
            <xsl:with-param name="word" select="$abbrev"/>
          </xsl:call-template>
        </xsl:otherwise>
      </xsl:choose>
    </xsl:if>
  </xsl:template>

  <xsl:template name="find-abbrev-in-title">
    <xsl:param name="title"/>
    <xsl:param name="word"/>
    <xsl:variable name="fullword" select="concat($word,' ')"/>
    <xsl:if test="string-length($word) > 1 and contains($title,$fullword)">
      <Word>
        <xsl:value-of select="$word"/>
      </Word>
      <xsl:text>&#10;</xsl:text>
    </xsl:if>
  </xsl:template>
</xsl:stylesheet>
