<?xml version="1.0" encoding="utf-8"?>
<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
                version="1.0">

  <!-- Use this script to generate a SQL script which inserts the full
       name and the standard abbreviation taken from the list of
       Pubmed Serials available at
       http://www.ncbi.nlm.nih.gov/entrez/citmatch_help.html#JournalLists
       into a RefDB reference database
       -->

  <xsl:output method="text"/>

  <xsl:template match="Serial">
    <xsl:choose>
      <xsl:when test="not(ISOAbbreviation)">
        <xsl:text>INSERT INTO t_periodical (periodical_name,periodical_abbrev) VALUES ('</xsl:text>
        <xsl:apply-templates select="Title"/>
        <xsl:text>','</xsl:text>
        <xsl:call-template name="add-periods-to-string">
          <xsl:with-param name="string" select="MedlineTA"/>
        </xsl:call-template>
        <xsl:text>');</xsl:text>
      </xsl:when>
      <xsl:otherwise>
        <xsl:text>INSERT INTO t_periodical (periodical_name,periodical_abbrev) VALUES ('</xsl:text>
        <xsl:apply-templates select="Title"/>
        <xsl:text>','</xsl:text>
        <xsl:call-template name="remove-space-after-dot">
          <xsl:with-param name="string" select="ISOAbbreviation"/>
        </xsl:call-template>
        <xsl:text>');</xsl:text>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>

  <!-- MedlineTA abbreviations do not use periods -->
  <!-- use document() to access the list of unabbreviated journal words, something like for-each select="document(words.xml)/set/word" -->
  <!-- implement a extractjournalwordsxml.xsl script to extract these words into a simple XML document -->
  <xsl:template name="add-periods-to-string">
    <xsl:param name="string"/>
    <xsl:if test="$string">
      <xsl:choose>
        <xsl:when test="contains($string, ' ')">
          <xsl:choose>
            <xsl:when test="document('journalwords.xml')/RefDBWordSet/Word[. = substring-before($string, ' ')]">
              <xsl:value-of select="substring-before($string, ' ')"/>
              <xsl:text> </xsl:text>
            </xsl:when>
            <xsl:otherwise>
              <xsl:value-of select="substring-before($string, ' ')"/>
              <xsl:text>.</xsl:text>
            </xsl:otherwise>
          </xsl:choose>
          <xsl:call-template name="add-periods-to-string">
            <xsl:with-param name="string" select="substring-after($string, ' ')"/>
          </xsl:call-template>
        </xsl:when>
        <xsl:otherwise>
          <xsl:choose>
            <xsl:when test="document('journalwords.xml')/RefDBWordSet/Word[. = $string]">
              <xsl:value-of select="$string"/>
            </xsl:when>
            <xsl:otherwise>
              <xsl:value-of select="$string"/>
              <xsl:text>.</xsl:text>
            </xsl:otherwise>
          </xsl:choose>
        </xsl:otherwise>
      </xsl:choose>
    </xsl:if>
  </xsl:template>

  <!-- ISOAbbreviations use '. ' as separators, turn into '.' -->
  <xsl:template name="remove-space-after-dot">
    <xsl:param name="string"/>
    <xsl:if test="$string">
      <xsl:choose>
        <xsl:when test="contains($string, '. ')">
          <xsl:value-of select="substring-before($string, ' ')"/>
          <xsl:call-template name="remove-space-after-dot">
            <xsl:with-param name="string" select="substring-after($string, ' ')"/>
          </xsl:call-template>
        </xsl:when>
        <xsl:otherwise>
          <xsl:value-of select="$string"/>
        </xsl:otherwise>
      </xsl:choose>
    </xsl:if>
  </xsl:template>

</xsl:stylesheet>
